/*
*******************************************************************************
* File:         E5270_TIS.c
* Description:  E5270A TIS Library C source file
* Revisions:    A.01.01   [ 3/1/2004 ]
* Language:     C
*
* (C) Copyright 2002-2004, Agilent Technologies Company, all rights reserved.
*******************************************************************************
*/

#include <stdio.h>
#include <math.h>
#include <time.h>
#include <float.h>
#include "E5270_TIS.h"

#define RETURN_PASS          0
#define RETURN_ERROR        -1

#define LIB_ERROR_MIN       10000
#define INVALID_PORT        (LIB_ERROR_MIN + 1)
#define INVALID_PORT_MSG    "Invalid port/pin number is specified."
#define INVALID_PIN         (LIB_ERROR_MIN + 2)
#define INVALID_PIN_MSG     "Invalid pin number is specified."
#define INVALID_RATIO       (LIB_ERROR_MIN + 3)
#define INVALID_RATIO_MSG   "The retio must be more then 0.01."
#define LIB_ERROR_MAX       10100

#define MAX_SMU            8
#define MAX_PIN           49
#define MAX_SWP_POINT   1001

#define INT_ERROR_PARA_VALUE  999999
#define SWEEP_DUMMY_DATA      9999999.99999
#define DUMMY_GNDU_CH         100
#define IGNORE_POWER_COMP     10000

enum { MM_NODATA, MM_SWEEP, MM_PULSE_SWEEP, MM_SWEEP_PULSE };
enum { ER_LIB, ER_PNP, ER_INST };

/* _ret shuld be rvalue */
#define CHECK_RET( _ret, _fname ) \
        { if ((_ret = checkRet( _ret, _fname )) != VI_SUCCESS ) return( RETURN_ERROR ); }
#define CHECK_CH( _ch, _fname ) \
        { if (!checkCh( _ch, _fname )) return( RETURN_ERROR ); }

/*== static variables =========================================*/
typedef struct { int point; 
                 int status[MAX_SWP_POINT * 2]; 
                 double time[MAX_SWP_POINT * 2]; } swp_stat_type;

typedef struct { int status; 
                 double time; } last_stat_type;

/* static global variable to keep E5270 internal settings */
static struct {
    ViSession vi;

    int    enable_error_detect;
    FILE  *fp_error_log;

    int    last_error;
    int    last_error_type;
    char   inst_error_mes[256];

    int    pin_port[MAX_PIN + 1];
    int    enable_timestamp;
    int    force_mode[MAX_SMU + 1];

    int    sweep_mm;
    int    sweep_ch;
    int    sweep_mode;
    double sweep_start;
    double sweep_stop;
    int    sweep_number;
    int    sweep_stop_mode;
    int    sweep_pcomp;

    int    sync_mm;
    int    sync_enable;
    int    sync_ch;
    double sync_start;
    double sync_stop;
    int    sync_pcomp;

    int    pbias_ch;
    int    pbias_mode;

    int    sweep_last_number;

    last_stat_type last[MAX_SMU + 1];
    swp_stat_type  swp[MAX_SMU + 1];

    int    tsr_flag;
} S_e5270;

/*== static functions =========================================*/
static void initE5270Data( void )
{
    int i;

    S_e5270.last_error = 0;
    S_e5270.last_error_type = ER_LIB;

    S_e5270.enable_timestamp = 0;
    for ( i = 0 ; i <= MAX_PIN ; i++ ) S_e5270.pin_port[i] = INT_ERROR_PARA_VALUE;
    for ( i = 0 ; i <= MAX_SMU ; i++ ) S_e5270.force_mode[i] = age5270_VF_MODE;

    S_e5270.sweep_mm = MM_NODATA;
    S_e5270.sweep_ch = 0;
    S_e5270.sweep_mode = 0;
    S_e5270.sweep_start = 0;
    S_e5270.sweep_stop = 0;
    S_e5270.sweep_number = 0;
    S_e5270.sweep_stop_mode = COMP_CONT;
    S_e5270.sweep_pcomp = 0;

    S_e5270.sync_mm = MM_NODATA;
    S_e5270.sync_enable = 0;
    S_e5270.sync_ch = 0;
    S_e5270.sync_start = 0;
    S_e5270.sync_stop = 0;
    S_e5270.sync_pcomp = 0;

    S_e5270.pbias_ch = 0;
    S_e5270.pbias_mode = 0;
    
    S_e5270.sweep_last_number = 0;

    for ( i = 0 ; i <= MAX_SMU ; i++ ) {
        S_e5270.last[i].status = 0;
        S_e5270.last[i].time = 0;
        S_e5270.swp[i].point = 0;
    }

    S_e5270.tsr_flag = 0;
}

static int checkCh( int ch, char *fname )
{
    if ( ch == INT_ERROR_PARA_VALUE ) {
        char dbuf[9];
        char tbuf[9];
        _strdate( dbuf );
        _strtime( tbuf );

        if ( S_e5270.fp_error_log != NULL ) {
            fprintf( S_e5270.fp_error_log, "%s %s : %s [Library Error %d]\n", 
                         dbuf, tbuf, fname, INVALID_PORT );
            fprintf( S_e5270.fp_error_log, "    %s\n", INVALID_PORT_MSG );
        }
        else {
            printf( "%s %s : %s [Library Error %d]\n", dbuf, tbuf, fname, INVALID_PORT );
            printf( "    %s\n", INVALID_PORT_MSG );
        }
        S_e5270.last_error = INVALID_PORT;
        S_e5270.last_error_type = ER_LIB;
        return( 0 );
    }
    return( 1 );
}

static int checkRet( int ret, char *fname )
{
    S_e5270.last_error = 0;
    if ( ret != VI_SUCCESS ) {

        int  errn;
        char errm[256];
        char dbuf[9];
        char tbuf[9];
        _strdate( dbuf );
        _strtime( tbuf );

        if ( ret == age5270_INSTR_ERROR_DETECTED ) {
            while ( 1 ) {
                age5270_error_query( S_e5270.vi, &errn, errm );
                if ( errn == 0 ) break;
                if ( errn == 227 ) {  /* Sweep measurement aborted */
                    ret = VI_SUCCESS;
                }
                else {
                    ret = errn;
                    S_e5270.last_error = ret;
                    S_e5270.last_error_type = ER_INST;
                    strcpy( S_e5270.inst_error_mes, errm );
                    if ( S_e5270.fp_error_log != NULL ) {
                        fprintf( S_e5270.fp_error_log, "%s %s : %s [INST Error %d]\n", 
                                     dbuf, tbuf, fname, errn );
                        fprintf( S_e5270.fp_error_log, "    %s\n", errm );
                    }
                    else {
                        printf( "%s %s : %s [INST Error %d]\n", 
                                     dbuf, tbuf, fname, errn );
                        printf( "    %s\n", errm );
                    }
                }
            }
        }
        else {
            S_e5270.last_error = ret;
            S_e5270.last_error_type = ER_PNP;
            errn = ret;
            age5270_error_message( S_e5270.vi, errn, errm );
            if ( S_e5270.fp_error_log != NULL ) {
                fprintf( S_e5270.fp_error_log, "%s %s : %s [PnP Error 0x%08x]\n",
                             dbuf, tbuf, fname, errn );
                fprintf( S_e5270.fp_error_log, "    %s\n", errm );
            }
            else {
                printf( "%s %s : %s [PnP Error 0x%08x]\n",
                             dbuf, tbuf, fname, errn );
                printf( "    %s\n", errm );
            }
        }
    }
    return( ret );
}

static int convPort( int port )
{
    switch ( port ) {
    case SMU1   : return(age5270_CH_1);
    case SMU2   : return(age5270_CH_2);
    case SMU3   : return(age5270_CH_3);
    case SMU4   : return(age5270_CH_4);
    case SMU5   : return(age5270_CH_5);
    case SMU6   : return(age5270_CH_6);
    case SMU7   : return(age5270_CH_7);
    case SMU8   : return(age5270_CH_8);
    case GNDU   : return(DUMMY_GNDU_CH);
    default     : return(INT_ERROR_PARA_VALUE);
    }
}

static int convPinPort( int pinport )
{
    switch ( pinport ) {
    case SMU1   : return(age5270_CH_1);
    case SMU2   : return(age5270_CH_2);
    case SMU3   : return(age5270_CH_3);
    case SMU4   : return(age5270_CH_4);
    case SMU5   : return(age5270_CH_5);
    case SMU6   : return(age5270_CH_6);
    case SMU7   : return(age5270_CH_7);
    case SMU8   : return(age5270_CH_8);
    case GNDU   : return(DUMMY_GNDU_CH);
    default     : if (( 1 <= pinport ) && ( pinport <= MAX_PIN ))
                      return( S_e5270.pin_port[pinport] );
                  else
                      return(INT_ERROR_PARA_VALUE);
    }
}

static int convAdc( int adc )
{
    switch ( adc ) {
    case PERCH_ADC : return(age5270_HSPEED_ADC);
    case REF_ADC   : return(age5270_HRESOLN_ADC);
    default        : return(INT_ERROR_PARA_VALUE);
    }
}

static int convAdcMode( int mode )
{
    switch ( mode ) {
    case INTEG_MANUAL  : return(age5270_INTEG_MANUAL);
    case INTEG_SHORT   : return(age5270_INTEG_AUTO);
    case INTEG_MEDIUM  : return(age5270_INTEG_PLC);
    case INTEG_LONG    : return(age5270_INTEG_PLC);
    default            : return(INT_ERROR_PARA_VALUE);
    }
}

static int convAutozero( int autozero )
{
    switch ( autozero ) {
    case AUTOZERO_OFF  : return(age5270_FLAG_OFF);
    case AUTOZERO_ON   : return(age5270_FLAG_ON);
    default            : return(INT_ERROR_PARA_VALUE);
    }
}
    
static int convOnOff( int on_off )
{
    switch ( on_off ) {
    case OFF  : return(age5270_FLAG_OFF);
    case ON   : return(age5270_FLAG_ON);
    default   : return(INT_ERROR_PARA_VALUE);
    }
}

static int convFilter( int on_off )
{
    switch ( on_off ) {
    case FILTER_OFF  : return(age5270_FLAG_OFF);
    case FILTER_ON   : return(age5270_FLAG_ON);
    default          : return(INT_ERROR_PARA_VALUE);
    }
}
    
static int convStopMode( int stop_mode )
{
    switch ( stop_mode ) {
    case COMP_CONT  : return(age5270_FLAG_OFF);
    case COMP_STOP  : return(age5270_FLAG_ON);
    default         : return(INT_ERROR_PARA_VALUE);
    }
}
    
static int convSwpMode( int mode )
{
    switch ( mode ) {
    case LINEAR_V      : return( age5270_SWP_VF_SGLLIN );
    case LINEAR_I      : return( age5270_SWP_IF_SGLLIN );
    case LINEAR_V_DBL  : return( age5270_SWP_VF_DBLLIN );
    case LINEAR_I_DBL  : return( age5270_SWP_IF_DBLLIN );
    case LOG_V         : return( age5270_SWP_VF_SGLLOG );
    case LOG_I         : return( age5270_SWP_IF_SGLLOG );
    case LOG_V_DBL     : return( age5270_SWP_VF_DBLLOG );
    case LOG_I_DBL     : return( age5270_SWP_IF_DBLLOG );
    default            : return(INT_ERROR_PARA_VALUE);
    }
}

static int getForceModeFromSwpMode( int mode )
{
    switch ( mode ) {
    case LINEAR_V      : 
    case LINEAR_V_DBL  : 
    case LOG_V         : 
    case LOG_V_DBL     : return( age5270_VF_MODE );
    case LINEAR_I      : 
    case LINEAR_I_DBL  : 
    case LOG_I         : 
    case LOG_I_DBL     : return( age5270_IF_MODE );
    default            : return( age5270_VF_MODE );
    }
}

static int convFrcMode( int mode )
{
    switch ( mode ) {
    case V_SOURCE  : return(age5270_VF_MODE);
    case I_SOURCE  : return(age5270_IF_MODE);
    default        : return(INT_ERROR_PARA_VALUE);
    }
}
    
static int convMeasMode( int mode )
{
    switch ( mode ) {
    case V_MEAS  : return(age5270_VM_MODE);
    case I_MEAS  : return(age5270_IM_MODE);
    default      : return(INT_ERROR_PARA_VALUE);
    }
}

static int convBdvInterval( int interval )
{
    switch ( interval ) {
    case BDV_INTVL_SHORT : return(age5270_SHORT_INTERVAL);
    case BDV_INTVL_LONG  : return(age5270_LONG_INTERVAL);
    default              : return(INT_ERROR_PARA_VALUE);
    }
}

static int convIleakInterval( int interval )
{
    switch ( interval ) {
    case ILEAK_INTVL_SHORT : return(age5270_SHORT_INTERVAL);
    case ILEAK_INTVL_LONG  : return(age5270_LONG_INTERVAL);
    default                : return(INT_ERROR_PARA_VALUE);
    }
}

static int convStatus( int status )
{
    if ( status == 0 ) return( NORMAL_MEAS );
    else if ( status & (1<<(age5270_STAT_OSC_BIT-1))) {
        return( DCS_OSC );
    }
    else if ( status & (1<<(age5270_STAT_OVF_BIT-1))) {
        return( DCS_OVERFLOW );
    }
    else if ( status & (1<<(age5270_STAT_TCC_BIT-1))) {
        return( DCS_COMP );
    }
    else if ( status & (1<<(age5270_STAT_OCC_BIT-1))) {
        return( DCS_COMP_OTHER );
    }
    else if ( status & (1<<(age5270_STAT_SAB_BIT-1))) {
        return( DCS_SWP_STOPPED );
    }
    else {
        return( status * -1 );
    }
}

static void calcuSweepStep( int mode, double start, double stop, int number, int point, 
                            double *sync )
{
    int i;
    double *ptr = sync;

    switch ( mode ) {
    case LINEAR_V      : 
    case LINEAR_I      : 
            if ( number == 1 ) {
                for ( i = 0 ; i < point ; i++ ) *ptr++ = start;
            }
            else {
                for ( i = 0 ; i < point ; i++ ) {
                    *ptr++ = start + i * ( stop - start ) / ( number - 1 );
                }
            }
            break;
    case LINEAR_V_DBL  : 
    case LINEAR_I_DBL  : 
            if ( number == 1 ) {
                for ( i = 0 ; i < point ; i++ ) *ptr++ = start;
            }
            else {
                for ( i = 0 ; i < point ; i++ ) {
                    if ( i < number ) 
                        *ptr++ = start + i * ( stop - start ) / ( number - 1 );
                    else 
                        *ptr++ = start + ( 2 * number - i - 1 ) * 
                                         ( stop - start ) / ( number - 1 );
                }
            }
            break;
    case LOG_V         : 
    case LOG_I         : 
            if ( number == 1 ) {
                for ( i = 0 ; i < point ; i++ ) *ptr++ = start;
            }
            else {
                for ( i = 0 ; i < point ; i++ ) {
                    *ptr++ = exp( log( stop / start ) / ( number - 1 ) * i ) * start;
                }
            }
            break;
    case LOG_V_DBL     : 
    case LOG_I_DBL     : 
            if ( number == 1 ) {
                for ( i = 0 ; i < point ; i++ ) *ptr++ = start;
            }
            else {
                for ( i = 0 ; i < point ; i++ ) {
                    if ( i < number ) 
                        *ptr++ = exp(log(stop/start)/(number-1)*i) * start;
                    else 
                        *ptr++ = exp(log(stop/start)/(number-1)*(2*number-i-1)) * start;
                }
            }
            break;
    }
}


/*== 4070 TIS incompatible functions =========================================*/
int open_E5270( char *instr_desc, int err_query, char *log_file )
{
    int ret = 0;

    initE5270Data();

    S_e5270.enable_error_detect = ( err_query == ERR_DETECT_ON );
    if ( log_file != NULL ) {
        S_e5270.fp_error_log = fopen( log_file, "a" );
        if ( S_e5270.fp_error_log == NULL ) {
            printf( "Error Logging File [%s] open error.\n", log_file );
        }
    }
    else {
        S_e5270.fp_error_log = NULL;
    }

    ret = age5270_init( instr_desc, VI_TRUE, VI_TRUE, &S_e5270.vi );
    CHECK_RET( ret, "open_E5270" );

    if ( err_query == ERR_DETECT_ON )
        ret = age5270_errorQueryDetect( S_e5270.vi, VI_TRUE );
    else
        ret = age5270_errorQueryDetect( S_e5270.vi, VI_FALSE );
    CHECK_RET( ret, "open_E5270" );

    return( RETURN_PASS );
}

int close_E5270( void )
{
    int ret = 0;
    if ( S_e5270.fp_error_log != NULL ) {
        fclose( S_e5270.fp_error_log );
        S_e5270.fp_error_log = NULL;
    }
    ret = age5270_close( S_e5270.vi );
    CHECK_RET( ret, "close_E5270" );
    return( RETURN_PASS );
}

ViSession vi_E5270( void )
{
    return( S_e5270.vi );
}


/*== 4070 TIS compatible functions =========================================*/
int init_system( void )
{
    int ret = 0;

    ret = age5270_dcl( S_e5270.vi );
    CHECK_RET( ret, "init_system" );
    ret = age5270_reset( S_e5270.vi );
    CHECK_RET( ret, "init_system" );
    initE5270Data();
    return( RETURN_PASS );
}

int set_adc( int adc, int mode, double value, int autozero )
{
#define INTEG_RESOLUTION 80.0e-6

    int ret = 0;
    int int_value;

    if ( adc == REF_ADC ) {
        switch ( mode ) {
        case INTEG_MANUAL  : if ( value >= 1.0 ) {
                                 int_value = (int)(value + 0.5);
                                 mode = INTEG_LONG;
                             }
                             else {
                                 int_value = (int)(value / INTEG_RESOLUTION + 0.5);
                                 if ( int_value == 0 ) int_value = 3;
                             }
                             break;
        case INTEG_SHORT   : int_value = (int)(value / INTEG_RESOLUTION + 0.5);
                             if ( int_value == 0 ) int_value = 6;
                             break;
        case INTEG_MEDIUM  : int_value = 1; /* 1PLC */
                             break;
        case INTEG_LONG    : int_value = (int)(value + 0.5);
                             if ( int_value == 0 ) int_value = 16;
                             break;
        default            : int_value = INT_ERROR_PARA_VALUE;
                             break;
        }
    }
    else {
        switch ( mode ) {
        case INTEG_MANUAL  : 
        case INTEG_SHORT   : int_value = (int)(value + 0.5);
                             if ( int_value == 0 ) int_value = 1;
                             break;
        case INTEG_MEDIUM  : int_value = 1; /* 1PLC */
                             break;
        case INTEG_LONG    : int_value = (int)(value + 0.5);
                             if ( int_value == 0 ) int_value = 16;
                             break;
        default            : int_value = INT_ERROR_PARA_VALUE;
                             break;
        }
    }
    ret = age5270_setAdc( S_e5270.vi, convAdc(adc), convAdcMode(mode), 
                                int_value, convAutozero(autozero) );
    CHECK_RET( ret, "set_adc" );
    return( RETURN_PASS );
}

int reset_timestamp( void )
{
    S_e5270.tsr_flag = 1;
    S_e5270.last_error = 0;
    return( RETURN_PASS );
}

int set_timestamp( int on_off )
{
    S_e5270.enable_timestamp = ( on_off != OFF );
    S_e5270.last_error = 0;
    return( RETURN_PASS );
}

int connect_pin( int port, int pin )
{
    int ret = 0;
    int i;
    int ch, ch2;
    int otherpin;

    if ( pin == 0 ) {
        switch ( port ) {
        case 0 :
            ret = age5270_setSwitch( S_e5270.vi, age5270_CH_ALL, age5270_FLAG_OFF );
            CHECK_RET( ret, "connect_pin" );
            for ( i = 0 ; i <= MAX_PIN ; i++ ) S_e5270.pin_port[i] = INT_ERROR_PARA_VALUE;
            for ( i = 0 ; i <= MAX_SMU ; i++ ) S_e5270.force_mode[i] = age5270_VF_MODE;
            break;
        case GNDU :
            for ( i = 0 ; i <= MAX_PIN ; i++ ) {
                if ( S_e5270.pin_port[i] == GNDU ) S_e5270.pin_port[i] = INT_ERROR_PARA_VALUE;
            }
            break;
        default :
            ch = convPort(port);
            CHECK_CH( ch, "connect_pin" );
            ret = age5270_setSwitch( S_e5270.vi, ch, age5270_FLAG_OFF );
            CHECK_RET( ret, "connect_pin" );
            S_e5270.force_mode[ch] = age5270_VF_MODE;
            for ( i = 0 ; i <= MAX_PIN ; i++ ) {
                if ( S_e5270.pin_port[i] == ch ) 
                    S_e5270.pin_port[i] = INT_ERROR_PARA_VALUE;
            }
        }
    }
    else {
        if (!(( 1 <= pin ) && ( pin <= MAX_PIN ))) {
            char dbuf[9];
            char tbuf[9];
            _strdate( dbuf );
            _strtime( tbuf );

            if ( S_e5270.fp_error_log != NULL ) {
                fprintf( S_e5270.fp_error_log, "%s %s : %s [Library Error %d]\n", 
                             dbuf, tbuf, "connect_pin", INVALID_PIN );
                fprintf( S_e5270.fp_error_log, "    %s\n", INVALID_PIN_MSG );
            }
            else {
                printf( "%s %s : %s [Library Error %d]\n", dbuf, tbuf, "connect_pin", INVALID_PIN );
                printf( "    %s\n", INVALID_PIN_MSG );
            }
            S_e5270.last_error = INVALID_PIN;
            S_e5270.last_error_type = ER_LIB;
            return( RETURN_ERROR );
        }

        switch ( port ) {
        case 0 :
            ch = convPinPort(pin);
            CHECK_CH( ch, "connect_pin" );
            otherpin = 0;
            for ( i = 0 ; i <= MAX_PIN ; i++ ) {
                if (( S_e5270.pin_port[i] == ch ) && ( i != pin )) {
                    otherpin = 1;
                    break;
                }
            }
            if ( otherpin == 0 ) {
                ret = age5270_setSwitch( S_e5270.vi, ch, age5270_FLAG_OFF );
                CHECK_RET( ret, "connect_pin" );
                S_e5270.pin_port[pin] = INT_ERROR_PARA_VALUE;
                S_e5270.force_mode[ch] = age5270_VF_MODE;
            }
            else {
                S_e5270.pin_port[pin] = INT_ERROR_PARA_VALUE;
            }
            break;
        case GNDU :
            if ( S_e5270.pin_port[pin] != INT_ERROR_PARA_VALUE ) {
                ch = S_e5270.pin_port[pin];
                otherpin = 0;
                for ( i = 0 ; i <= MAX_PIN ; i++ ) {
                    if (( S_e5270.pin_port[i] == ch ) && ( i != pin )) {
                        otherpin = 1;
                        break;
                    }
                }
                if ( otherpin == 0 ) {
                    ret = age5270_setSwitch( S_e5270.vi, ch, age5270_FLAG_OFF );
                    CHECK_RET( ret, "connect_pin" );
                    S_e5270.force_mode[ch] = age5270_VF_MODE;
                }
            }
            S_e5270.pin_port[pin] = convPort(port);
            break;
        default :
            ch = convPort(port);
            CHECK_CH( ch, "connect_pin" );
            if (( S_e5270.pin_port[pin] != INT_ERROR_PARA_VALUE ) &&
                ( S_e5270.pin_port[pin] != DUMMY_GNDU_CH )) {
                ch2 = S_e5270.pin_port[pin];
                otherpin = 0;
                for ( i = 0 ; i <= MAX_PIN ; i++ ) {
                    if (( S_e5270.pin_port[i] == ch2 ) && ( i != pin )) {
                        otherpin = 1;
                        break;
                    }
                }
                if ( otherpin == 0 ) {
                    ret = age5270_setSwitch( S_e5270.vi, ch2, age5270_FLAG_OFF );
                    CHECK_RET( ret, "connect_pin" );
                    S_e5270.force_mode[ch2] = age5270_VF_MODE;
                }
            }
            ret = age5270_setSwitch( S_e5270.vi, ch, age5270_FLAG_ON );
            CHECK_RET( ret, "connect_pin" );
            S_e5270.pin_port[pin] = ch;
        }
    }
    return( RETURN_PASS );
}

int disconnect_all( void )
{
    int i;
    int ret = 0;

    ret = age5270_setSwitch( S_e5270.vi, age5270_CH_ALL, age5270_FLAG_OFF );
    CHECK_RET( ret, "disconnect_all" );
    for ( i = 0 ; i <= MAX_PIN ; i++ ) {
        S_e5270.pin_port[i] = INT_ERROR_PARA_VALUE;
        S_e5270.force_mode[i] = age5270_VF_MODE;
    }
    return( RETURN_PASS );
}

int disable_port( int port )
{
    int i;
    int ret = 0;
    int ch;

    if ( port == 0 ) {
        ret = age5270_zeroOutput( S_e5270.vi, age5270_CH_ALL );
    }
    else {
        ch = convPinPort(port);
        CHECK_CH( ch, "disable_port" );
        ret = age5270_zeroOutput( S_e5270.vi, ch );
    }
    CHECK_RET( ret, "disable_port" );
    if ( port == 0 ) {
        for ( i = 0 ; i <= MAX_SMU ; i++ ) S_e5270.force_mode[i] = age5270_VF_MODE;
        /* Disable Pbias port */
        if ( S_e5270.sweep_mm == MM_SWEEP_PULSE ) S_e5270.sweep_mm = MM_SWEEP;
        S_e5270.pbias_ch = 0;
        S_e5270.pbias_mode = 0;
    }
    else {
        S_e5270.force_mode[ch] = age5270_VF_MODE;
        if ( ch == S_e5270.pbias_ch ) {
            /* Disable Pbias port */
            if ( S_e5270.sweep_mm == MM_SWEEP_PULSE ) S_e5270.sweep_mm = MM_SWEEP;
            S_e5270.pbias_ch = 0;
            S_e5270.pbias_mode = 0;
        }
    }
    return( RETURN_PASS );
}

int set_smu_ch( int pinport, int adc, int filter )
{
    int ret = 0;
    
    if ( pinport == 0 ) {
        ret = age5270_setAdcType( S_e5270.vi, age5270_CH_ALL, convAdc(adc) );
        CHECK_RET( ret, "set_smu_ch" );
        ret = age5270_setFilter( S_e5270.vi, age5270_CH_ALL, convFilter(filter) );
        CHECK_RET( ret, "set_smu_ch" );
    }
    else {
        int ch = convPinPort( pinport );

        CHECK_CH( ch, "set_smu_ch" );
        if (ch == DUMMY_GNDU_CH) return( RETURN_PASS );
        ret = age5270_setAdcType( S_e5270.vi, ch, convAdc(adc) );
        CHECK_RET( ret, "set_smu_ch" );
        ret = age5270_setFilter( S_e5270.vi, ch, convFilter(filter) );
        CHECK_RET( ret, "set_smu_ch" );
    }
    return( RETURN_PASS );
}

int force_v( int pinport, double voltage, double range, double compliance )
{
    int ret = 0;
    int ch = convPinPort( pinport );

    CHECK_CH( ch, "force_v" );
    ret = age5270_force( S_e5270.vi, ch, age5270_VF_MODE, 
                         range, voltage, compliance, age5270_POL_AUTO );
    CHECK_RET( ret, "force_v" );
    S_e5270.force_mode[ch] = age5270_VF_MODE;
    if ( S_e5270.tsr_flag == 1 ) {
        ret = age5270_resetTimestamp( S_e5270.vi );
        CHECK_RET( ret, "force_v" );
        S_e5270.tsr_flag = 0;
    }
    return( RETURN_PASS );
}

int force_i( int pinport, double current, double range, double compliance )
{
    int ret = 0;
    int ch = convPinPort( pinport );

    CHECK_CH( ch, "force_i" );
    ret = age5270_force( S_e5270.vi, ch, age5270_IF_MODE, 
                         range, current, compliance, age5270_POL_AUTO );
    CHECK_RET( ret, "force_i" );
    S_e5270.force_mode[ch] = age5270_IF_MODE;
    if ( S_e5270.tsr_flag == 1 ) {
        ret = age5270_resetTimestamp( S_e5270.vi );
        CHECK_RET( ret, "force_i" );
        S_e5270.tsr_flag = 0;
    }
    return( RETURN_PASS );
}

int set_iv( int pinport, int sweep_mode, double range, double start, double stop,
            int number, double hold, double delay, 
            double compliance, double power_compliance, int stop_mode )
{
    int ret = 0;
    int ch = convPinPort( pinport );

    CHECK_CH( ch, "set_iv" );
    if ( power_compliance <= DBL_EPSILON ) power_compliance = IGNORE_POWER_COMP;
    ret = age5270_setIv( S_e5270.vi, ch, convSwpMode(sweep_mode),
                         range, start, stop, number, hold, delay, 0.0 /*s_delay*/, 
                         compliance, power_compliance );
    CHECK_RET( ret, "set_iv" );
    S_e5270.sweep_mm = MM_SWEEP;
    S_e5270.sweep_ch = ch;
    S_e5270.sweep_mode = sweep_mode;
    S_e5270.sweep_start = start;
    S_e5270.sweep_stop = stop;
    S_e5270.sweep_number = number;
    S_e5270.sweep_stop_mode = stop_mode;
    S_e5270.sweep_pcomp = ( power_compliance > 100 ? 0 : 1 );
    S_e5270.sync_mm = MM_SWEEP;
    S_e5270.sync_enable = 0;
    ret = age5270_stopMode( S_e5270.vi, convStopMode(stop_mode), age5270_LAST_START );
    CHECK_RET( ret, "set_iv" );
    return( RETURN_PASS );
}

int set_piv( int pinport, int sweep_mode, double range, 
             double base, double start, double stop,
             int number, double width, double period, double hold,
             double compliance, int stop_mode )
{
    int ret = 0;
    int ch = convPinPort( pinport );

    CHECK_CH( ch, "set_piv" );
    ret = age5270_setPiv( S_e5270.vi, ch, convSwpMode(sweep_mode),
                         range, base, start, stop, number, hold, width, period,
                         compliance );
    CHECK_RET( ret, "set_piv" );
    S_e5270.sweep_mm = MM_PULSE_SWEEP;
    S_e5270.sweep_ch = ch;
    S_e5270.sweep_mode = sweep_mode;
    S_e5270.sweep_start = start;
    S_e5270.sweep_stop = stop;
    S_e5270.sweep_number = number;
    S_e5270.sweep_stop_mode = stop_mode;
    S_e5270.sweep_pcomp = 0;
    S_e5270.sync_mm = MM_PULSE_SWEEP;
    S_e5270.sync_enable = 0;
    ret = age5270_stopMode( S_e5270.vi, convStopMode(stop_mode), age5270_LAST_START );
    CHECK_RET( ret, "set_piv" );
    return( RETURN_PASS );
}

int set_pbias( int pinport, int mode, double range, double base, double peak, 
               double width, double period, double hold, double compliance)
{
    int ret = 0;
    int ch = convPinPort( pinport );

    CHECK_CH( ch, "set_pbias" );
    ret = age5270_setPbias( S_e5270.vi, ch, convFrcMode(mode),
                         range, base, peak, width, period, hold, compliance );
    CHECK_RET( ret, "set_pbias" );
    if ( S_e5270.sweep_mm == MM_SWEEP ) S_e5270.sweep_mm = MM_SWEEP_PULSE;
    S_e5270.pbias_ch = ch;
    S_e5270.pbias_mode = mode;
    return( RETURN_PASS );
}

int set_sync( int pinport, int mode, double offset, double ratio, 
              double compliance, double power_compliance )
{
    int ret = 0;
    int frc_mode;
    double range, start, stop;
    int ch = convPinPort( pinport );

    CHECK_CH( ch, "set_sync" );
    if (( S_e5270.sync_mm == MM_SWEEP ) ||
        ( S_e5270.sync_mm == MM_PULSE_SWEEP ) ||
        ( S_e5270.sync_mm == MM_SWEEP_PULSE )) {

#define MIN_RATIO 0.01

        if ( ratio < MIN_RATIO ) {
            char dbuf[9];
            char tbuf[9];
            _strdate( dbuf );
            _strtime( tbuf );

            if ( S_e5270.fp_error_log != NULL ) {
                fprintf( S_e5270.fp_error_log, "%s %s : %s [Library Error %d]\n", 
                             dbuf, tbuf, "set_sync", INVALID_RATIO );
                fprintf( S_e5270.fp_error_log, "    %s\n", INVALID_RATIO_MSG );
            }
            else {
                printf( "%s %s : %s [Library Error %d]\n", 
                             dbuf, tbuf, "set_sync", INVALID_RATIO );
                printf( "    %s\n", INVALID_RATIO_MSG );
            }
            S_e5270.last_error = INVALID_RATIO;
            S_e5270.last_error_type = ER_LIB;
            return( RETURN_ERROR );
        }

        switch ( S_e5270.sweep_mode ) {
        case LOG_V :
        case LOG_V_DBL :
        case LOG_I :
        case LOG_I_DBL :
            start = S_e5270.sweep_start * ratio * (mode==POS_SYNC?1:-1);
            stop  = S_e5270.sweep_stop  * ratio * (mode==POS_SYNC?1:-1);
            break;
        default :
            start = S_e5270.sweep_start * ratio * (mode==POS_SYNC?1:-1) + offset;
            stop  = S_e5270.sweep_stop  * ratio * (mode==POS_SYNC?1:-1) + offset;
            break;
        }
        frc_mode = getForceModeFromSwpMode( S_e5270.sweep_mode );
        range = (fabs(start) > fabs(stop) ? fabs(start) : fabs(stop));

        if ( power_compliance <= DBL_EPSILON ) power_compliance = IGNORE_POWER_COMP;
        ret = age5270_setSweepSync( S_e5270.vi, ch, frc_mode, range, start, stop, 
                                compliance, power_compliance );
        CHECK_RET( ret, "set_sync" );
        S_e5270.sync_enable = 1;
        S_e5270.sync_ch = ch;
        S_e5270.sync_start = start;
        S_e5270.sync_stop = stop;
        S_e5270.sync_stop = stop;
        S_e5270.sync_pcomp = ( power_compliance > 100 ? 0 : 1 );
    }
    return( RETURN_PASS );
}

int measure_v( int pinport, double *voltage, double range)
{
    int    ret = 0;
    int    status;
    double time;
    int    ch = convPinPort( pinport );

    CHECK_CH( ch, "measure_v" );
    if ( S_e5270.enable_timestamp ) {
        ret = age5270_spotMeas( S_e5270.vi, ch, age5270_VM_MODE, range, 
                                voltage, &status,
                                &time );
    }
    else {
        ret = age5270_spotMeas( S_e5270.vi, ch, age5270_VM_MODE, range, 
                                voltage, &status, NULL );
        time = 0.0;
    }
    CHECK_RET( ret, "measure_v" );
    S_e5270.last[ch].status = convStatus(status);
    S_e5270.last[ch].time = time;
    return( RETURN_PASS );
}

int measure_i( int pinport, double *current, double range)
{
    int    ret = 0;
    int    status;
    double time;
    int    ch = convPinPort( pinport );

    CHECK_CH( ch, "measure_i" );
    if ( S_e5270.enable_timestamp ) {
        ret = age5270_spotMeas( S_e5270.vi, ch, age5270_IM_MODE, range, 
                                current, &status,
                                &time );
    }
    else {
        ret = age5270_spotMeas( S_e5270.vi, ch, age5270_IM_MODE, range, 
                                current, &status, NULL );
        time = 0.0;
    }
    CHECK_RET( ret, "measure_i" );
    S_e5270.last[ch].status = convStatus(status);
    S_e5270.last[ch].time = time;
    return( RETURN_PASS );
}

int measure_vt( int pinport, double *voltage, double range, double *time_stamp)
{
    int    ret = 0;
    int    status;
    int    ch = convPinPort( pinport );

    CHECK_CH( ch, "measure_vt" );
    if ( S_e5270.enable_timestamp ) {
        ret = age5270_spotMeas( S_e5270.vi, ch, age5270_VM_MODE, range, voltage, 
                                &status, time_stamp );
    }
    else {
        ret = age5270_spotMeas( S_e5270.vi, ch, age5270_VM_MODE, range, voltage, 
                                &status, NULL );
        *time_stamp = 0.0;
    }
    CHECK_RET( ret, "measure_vt" );
    S_e5270.last[ch].status = convStatus(status);
    S_e5270.last[ch].time = *time_stamp;
    return( RETURN_PASS );
}

int measure_it( int pinport, double *current, double range, double *time_stamp)
{
    int    ret = 0;
    int    status;
    int    ch = convPinPort( pinport );

    CHECK_CH( ch, "measure_it" );
    if ( S_e5270.enable_timestamp ) {
        ret = age5270_spotMeas( S_e5270.vi, ch, age5270_IM_MODE, range, current, 
                                &status, time_stamp );
    }
    else {
        ret = age5270_spotMeas( S_e5270.vi, ch, age5270_IM_MODE, range, current, 
                                &status, NULL );
        *time_stamp = 0.0;
    }
    CHECK_RET( ret, "measure_it" );
    S_e5270.last[ch].status = convStatus(status);
    S_e5270.last[ch].time = *time_stamp;
    return( RETURN_PASS );
}

int measure_m( int n, int *pinports, double *meas_vals, double *ranges, double *time_stamps)
{
    int ret = 0;
    int channel[MAX_SMU + 1];
    int status[MAX_SMU + 1];
    int mode[MAX_SMU + 1];
    int i;
    for ( i = 0 ; i < n ; i++ ) {
        channel[i] = convPinPort(pinports[i]);
        CHECK_CH( channel[i], "measure_m" );
        if (( age5270_CH_1 <= channel[i] ) && ( channel[i] <= age5270_CH_8 )) {
            mode[i] = (S_e5270.force_mode[channel[i]]==age5270_VF_MODE
                           ?age5270_IF_MODE:age5270_VF_MODE);
        }
        else {
            mode[i] = age5270_VF_MODE;
        }
    }
    channel[i] = 0;
    mode[i] = 0;
    if ( S_e5270.enable_timestamp ) {
        ret = age5270_measureM( S_e5270.vi, channel, mode, ranges,
                                meas_vals, status, time_stamps);
    }
    else {
        ret = age5270_measureM( S_e5270.vi, channel, mode, ranges,
                                meas_vals, status, NULL);
        for ( i = 0 ; i < n ; i++ ) time_stamps[i] = 0.0;
    }
    CHECK_RET( ret, "measure_m" );
    for ( i = 0 ; i < n ; i++ ) {
        S_e5270.last[channel[i]].status = convStatus(status[i]);
        S_e5270.last[channel[i]].time   = time_stamps[i];
    }
    return( RETURN_PASS );
}

int measure_p( int pinport, int mode, double *value, double range )
{
    int    ret = 0;
    int    status;
    double data;
    double time;
    int    ch;
    char   buf[100];

    if ( pinport != 0 ) ch = convPinPort( pinport );
    else {
        unsigned int i;
        int found;
        ret = age5270_cmdString_Q( S_e5270.vi, "UNT?", 100, buf );
        ch = 1;
        found = 0;
        for ( i = 0 ; i < strlen( buf ) ; i++ ) {
            switch ( buf[i] ) {
            case ',' :
            case '0' : break;
            case ';' : ch++; break;
            default  : found = 1; break;
            }
            if ( found != 0 ) break;
        }
    }

    CHECK_CH( ch, "measure_p" );
    if ( S_e5270.enable_timestamp ) {
        ret = age5270_measureP( S_e5270.vi, ch, convMeasMode(mode), 
                                range, &data, &status, &time );
    }
    else {
        ret = age5270_measureP( S_e5270.vi, ch, convMeasMode(mode), 
                                range, &data, &status, NULL );
        time = 0.0;
    }
    CHECK_RET( ret, "measure_p" );
    if ( pinport != 0 ) {
        *value = data;
        S_e5270.last[ch].status = convStatus(status);
        S_e5270.last[ch].time = time;
    }
    return( RETURN_PASS );
}

int port_status( int pinport, int *status )
{
    int ret = 0;
    int ch = convPinPort( pinport );

    CHECK_CH( ch, "port_status" );
    if (( age5270_CH_1 <= ch ) && ( ch <= age5270_CH_8 )) {
        *status = S_e5270.last[ch].status;
    }
    else {
        *status = NORMAL_MEAS;
    }
    return( RETURN_PASS );
}

int port_status_t( int pinport, int *status, double *time_stamp )
{
    int ret = 0;
    int ch = convPinPort( pinport );

    CHECK_CH( ch, "port_status_t" );
    if (( age5270_CH_1 <= ch ) && ( ch <= age5270_CH_8 )) {
        *status     = S_e5270.last[ch].status;
        *time_stamp = S_e5270.last[ch].time;
    }
    else {
        *status = NORMAL_MEAS;
        *time_stamp = 0.0;
    }
    return( RETURN_PASS );
}

int sweep_iv( int pinport, int mode, double range, 
              double *measure, double *source, double *sync)
{
    int    ret = 0;
    int    i;
    int    ch = convPinPort(pinport);
    int    point;
    int    status[MAX_SWP_POINT * 2];
    double time[MAX_SWP_POINT * 2];

    CHECK_CH( ch, "sweep_iv" );
    switch ( S_e5270.sweep_mm ) {
    case MM_SWEEP :
            if ( S_e5270.enable_timestamp ) {
                ret = age5270_sweepIv( S_e5270.vi, ch, convMeasMode(mode), 
                                   range, &point, source, measure, status, time );
                CHECK_RET( ret, "sweep_iv" );
                S_e5270.swp[ch].point = point;
                for ( i = 0 ; i < S_e5270.swp[ch].point ; i++ ) {
                    S_e5270.swp[ch].status[i] = convStatus(status[i]);
                    S_e5270.swp[ch].time[i] = time[i];
                }
            }
            else {
                ret = age5270_sweepIv( S_e5270.vi, ch, convMeasMode(mode), 
                                   range, &point, source, measure, status, NULL );
                CHECK_RET( ret, "sweep_iv" );
                S_e5270.swp[ch].point = point;
                for ( i = 0 ; i < S_e5270.swp[ch].point ; i++ ) {
                    S_e5270.swp[ch].status[i] = convStatus(status[i]);
                    S_e5270.swp[ch].time[i] = 0.0;
                }
            }
            break;
    case MM_PULSE_SWEEP :
            if ( S_e5270.enable_timestamp ) {
                ret = age5270_sweepPiv( S_e5270.vi, ch, convMeasMode(mode), 
                                   range, &point, source, measure, status, time );
                CHECK_RET( ret, "sweep_iv" );
                S_e5270.swp[ch].point = point;
                for ( i = 0 ; i < S_e5270.swp[ch].point ; i++ ) {
                    S_e5270.swp[ch].status[i] = convStatus(status[i]);
                    S_e5270.swp[ch].time[i] = time[i];
                }
            }
            else {
                ret = age5270_sweepPiv( S_e5270.vi, ch, convMeasMode(mode), 
                                   range, &point, source, measure, status, NULL );
                CHECK_RET( ret, "sweep_iv" );
                S_e5270.swp[ch].point = point;
                for ( i = 0 ; i < S_e5270.swp[ch].point ; i++ ) {
                    S_e5270.swp[ch].status[i] = convStatus(status[i]);
                    S_e5270.swp[ch].time[i] = 0.0;
                }
            }
            break;
    case MM_SWEEP_PULSE :
            if ( S_e5270.enable_timestamp ) {
                ret = age5270_sweepPbias( S_e5270.vi, ch, convMeasMode(mode), 
                                   range, &point, source, measure, status, time );
                CHECK_RET( ret, "sweep_iv" );
                S_e5270.swp[ch].point = point;
                for ( i = 0 ; i < S_e5270.swp[ch].point ; i++ ) {
                    S_e5270.swp[ch].status[i] = convStatus(status[i]);
                    S_e5270.swp[ch].time[i] = time[i];
                }
            }
            else {
                ret = age5270_sweepPbias( S_e5270.vi, ch, convMeasMode(mode), 
                                   range, &point, source, measure, status, NULL );
                CHECK_RET( ret, "sweep_iv" );
                S_e5270.swp[ch].point = point;
                for ( i = 0 ; i < S_e5270.swp[ch].point ; i++ ) {
                    S_e5270.swp[ch].status[i] = convStatus(status[i]);
                    S_e5270.swp[ch].time[i] = 0.0;
                }
            }
            S_e5270.force_mode[S_e5270.pbias_ch]
                = getForceModeFromSwpMode( S_e5270.pbias_mode );
            break;
    }
    if (( sync != NULL ) && S_e5270.sync_enable ) {
        calcuSweepStep( S_e5270.sweep_mode, 
                        S_e5270.sync_start, S_e5270.sync_stop, 
                        S_e5270.sweep_number, S_e5270.swp[ch].point,
                        sync );
    }
    switch ( S_e5270.sweep_mode ) {
    case LINEAR_V_DBL :
    case LINEAR_I_DBL :
    case LOG_V_DBL :
    case LOG_I_DBL :
        S_e5270.sweep_last_number = S_e5270.sweep_number * 2;
        break;
    default :
        S_e5270.sweep_last_number = S_e5270.sweep_number;
        break;
    }
    if (( S_e5270.sweep_stop_mode == COMP_STOP ) ||
        ( S_e5270.sweep_pcomp == 1 ) ||
        (( S_e5270.sync_enable ) && ( S_e5270.sync_pcomp == 1 ))) { 
        /* Auto Stop Enable */
        int auto_stopped = 0;
        for ( i = 0 ; i < S_e5270.sweep_last_number ; i++ ) {
            if ( auto_stopped == 1 ) {
                *(measure + i) = SWEEP_DUMMY_DATA;
                *(source + i) = 0.0;
                if (( sync != NULL ) && S_e5270.sync_enable ) {
                    *(sync + i) = 0.0;
                }
            }
            if ( S_e5270.swp[ch].status[i] != NORMAL_MEAS ) {
                if (( auto_stopped == 0 ) && ( i < S_e5270.sweep_last_number - 1 )) {
                    S_e5270.swp[ch].point = i+1;
                }
                auto_stopped = 1;
            }
        }
    }
    S_e5270.last[ch].status = S_e5270.swp[ch].status[S_e5270.swp[ch].point];
    S_e5270.last[ch].time   = S_e5270.swp[ch].time[S_e5270.swp[ch].point];
    S_e5270.force_mode[ch] = getForceModeFromSwpMode( S_e5270.sweep_mode );
    if ( S_e5270.sync_enable ) {
        S_e5270.force_mode[S_e5270.sync_ch]
                = getForceModeFromSwpMode( S_e5270.sweep_mode );
    }

    return( RETURN_PASS );
}

int sweep_miv( int n, int *pinports, double *ranges, 
               double *meas_vals, double *source, double *sync)
{
    int    ret = 0;
    int    point;
    int    channel[MAX_SMU + 1];
    int    mode[MAX_SMU + 1];
    double measure[(MAX_SMU + 1) * MAX_SWP_POINT * 2];
    int    status[(MAX_SMU + 1) * MAX_SWP_POINT * 2];
    double time[(MAX_SMU + 1) * MAX_SWP_POINT * 2];
    int    dum_sweep_last_number;
    int    i, j;

    switch ( S_e5270.sweep_mode ) {
    case LINEAR_V_DBL :
    case LINEAR_I_DBL :
    case LOG_V_DBL :
    case LOG_I_DBL :
        dum_sweep_last_number = S_e5270.sweep_number * 2;
        break;
    default :
        dum_sweep_last_number = S_e5270.sweep_number;
        break;
    }

    for ( i = 0 ; i < n ; i++ ) {
        channel[i] = convPinPort(pinports[i]);
        CHECK_CH( channel[i], "sweep_miv" );
        if (( channel[i] == S_e5270.sweep_ch ) ||
            ( S_e5270.sync_enable && ( channel[i] == S_e5270.sync_ch ))) {
            if ( getForceModeFromSwpMode( S_e5270.sweep_mode ) == age5270_VF_MODE )
                mode[i] = age5270_IM_MODE;
            else 
                mode[i] = age5270_VM_MODE;
        }
        else {
            mode[i] = (S_e5270.force_mode[channel[i]]==age5270_VF_MODE
                       ?age5270_IF_MODE:age5270_VF_MODE);
        }
    }
    channel[i] = 0;
    mode[i] = 0;

    if ( S_e5270.enable_timestamp ) {
        ret = age5270_sweepMiv( S_e5270.vi, channel, mode, ranges,
                           &point, source, measure, status, time );
        CHECK_RET( ret, "sweep_miv" );
        for ( i = 0 ; i < n ; i++ ) {
            S_e5270.swp[channel[i]].point = point;
            for ( j = 0 ; j < point ; j++ ) {
                *(meas_vals + i * dum_sweep_last_number + j) = measure[n * j + i];
                S_e5270.swp[channel[i]].status[j] = convStatus(status[n * j + i]);
                S_e5270.swp[channel[i]].time[j] = time[n * j + i];
            }
        }
    }
    else {
        ret = age5270_sweepMiv( S_e5270.vi, channel, mode, ranges, 
                           &point, source, measure, status, NULL );
        CHECK_RET( ret, "sweep_miv" );
        for ( i = 0 ; i < n ; i++ ) {
            S_e5270.swp[channel[i]].point = point;
            for ( j = 0 ; j < point ; j++ ) {
                *(meas_vals + i * dum_sweep_last_number + j) = measure[n * j + i];
                S_e5270.swp[channel[i]].status[j] = convStatus(status[n * j + i]);
                S_e5270.swp[channel[i]].time[j] = 0.0;
            }
        }
    }
    if (( sync != NULL ) && S_e5270.sync_enable ) {
        calcuSweepStep( S_e5270.sweep_mode, 
                        S_e5270.sync_start, S_e5270.sync_stop, 
                        S_e5270.sweep_number, point,
                        sync );
    }
    switch ( S_e5270.sweep_mode ) {
    case LINEAR_V_DBL :
    case LINEAR_I_DBL :
    case LOG_V_DBL :
    case LOG_I_DBL :
        S_e5270.sweep_last_number = S_e5270.sweep_number * 2;
        break;
    default :
        break;
    }
    S_e5270.sweep_last_number = dum_sweep_last_number;
    if (( S_e5270.sweep_stop_mode == COMP_STOP ) ||
        ( S_e5270.sweep_pcomp == 1 ) ||
        (( S_e5270.sync_enable ) && ( S_e5270.sync_pcomp == 1 ))) { 
        /* Auto Stop Enable */
        int auto_stopped = 0;
        for ( i = 0 ; i < S_e5270.sweep_last_number ; i++ ) {
            if ( auto_stopped == 1 ) {
                for ( j = 0 ; j < n ; j++ ) 
                    *(meas_vals + S_e5270.sweep_last_number * j + i) = SWEEP_DUMMY_DATA;
                *(source + i) = 0.0;
                if (( sync != NULL ) && S_e5270.sync_enable ) {
                    *(sync + i) = 0.0;
                }
            }
            for ( j = 0 ; j < n ; j++ ) {
                if ( S_e5270.swp[channel[j]].status[i] != NORMAL_MEAS ) {
                    if (( auto_stopped == 0 ) && ( i < S_e5270.sweep_last_number - 1 )) {
                        int k;
                        for ( k = 0 ; k < n ; k++ ) S_e5270.swp[channel[k]].point = i+1;
                    }
                    auto_stopped = 1;
                }
            }
        }
    }
    for ( i = 0 ; i < n ; i++ ) {
        S_e5270.last[channel[i]].status 
            = S_e5270.swp[channel[i]].status[S_e5270.swp[channel[i]].point];
        S_e5270.last[channel[i]].time   
            = S_e5270.swp[channel[i]].time[S_e5270.swp[channel[i]].point];
    }
    S_e5270.force_mode[S_e5270.sweep_ch]
            = getForceModeFromSwpMode( S_e5270.sweep_mode );
    if ( S_e5270.sync_enable ) {
        S_e5270.force_mode[S_e5270.sync_ch]
                = getForceModeFromSwpMode( S_e5270.sweep_mode );
    }

    return( RETURN_PASS );
}

int status_miv( int n, int *ports, int *statuses, double *times)
{
    int i,j;
    int ch;

    for ( i = 0 ; i < n ; i++ ) {
        ch = convPinPort(ports[i]);
        CHECK_CH( ch, "status_miv" );
        for ( j = 0 ; j < S_e5270.sweep_last_number ; j++ ) {
            if (( age5270_CH_1 <= ch ) && ( ch <= age5270_CH_8 )) {
                if ( j < S_e5270.swp[ch].point ) {
                    *(statuses + i * S_e5270.sweep_last_number + j) = S_e5270.swp[ch].status[j];
                    *(times + i * S_e5270.sweep_last_number + j)    = S_e5270.swp[ch].time[j];
                }
                else {
                    *(statuses + i * S_e5270.sweep_last_number + j) = DCS_SWP_STOPPED;
                    *(times + i * S_e5270.sweep_last_number + j)    = 0.0;
                }
            }
            else {
                *(statuses + i * S_e5270.sweep_last_number + j) = DCS_SWP_STOPPED;
                *(times + i * S_e5270.sweep_last_number + j)    = 0.0;
            }
        }
    }

    S_e5270.last_error = 0;
    return( RETURN_PASS );
}

int set_bdv( int port, double range, double start, double stop, double current, 
                    double hold, double delay )
{
    int ret = 0;
    int ch = convPinPort(port);

    CHECK_CH( ch, "set_bdv" );
    ret = age5270_setBdv( S_e5270.vi, convPinPort(port), range, start, stop, current, 
                          hold, delay );
    CHECK_RET( ret, "set_bdv" );
    return( RETURN_PASS );
}

int measure_bdv( double *voltage, int *status, int interval )
{
    int ret = 0;
    int st;

    ret = age5270_measureBdv( S_e5270.vi, convBdvInterval(interval), voltage, &st );
    CHECK_RET( ret, "measure_bdv" );

    if ( st == 0 )                                *status = BDV_CURRENT_LOW;
    else if ( st & (1<<(age5270_STAT_OSC_BIT-1))) *status = BDV_OSC;
    else if ( st & (1<<(age5270_STAT_OVF_BIT-1))) *status = BDV_OVERFLOW;
    else if ( st & (1<<(age5270_STAT_TCC_BIT-1))) *status = BDV_NORMAL;
    else if ( st & (1<<(age5270_STAT_OCC_BIT-1))) *status = BDV_COMP_OTHER;
    else if ( st & (1<<(age5270_STAT_SAB_BIT-1))) *status = BDV_TOO_SLOW;
    else if ( st & (1<<(age5270_STAT_SNF_BIT-1))) *status = BDV_TIMEOUT;
    else                                          *status = BDV_CURRENT_LOW;
    return( RETURN_PASS );
}

int set_ileak( int port, double range, double voltage, double compliance, 
                      double start, double hold, double delay )
{
    int ret = 0;
    int ch = convPinPort(port);

    CHECK_CH( ch, "set_ileak" );
    ret = age5270_setIleak( S_e5270.vi, convPinPort(port), range, voltage, compliance, 
                            start, hold, delay );
    CHECK_RET( ret, "set_ileak" );
    return( RETURN_PASS );
}

int measure_ileak( int port, double *current, int *status, int interval )
{
    int ret = 0;
    int st;

    ret = age5270_measureIleak( S_e5270.vi, convPinPort(port), 
                                convBdvInterval(interval), 
                                current, &st );
    CHECK_RET( ret, "measure_ileak" );

    if ( st == 0 )                                *status = ILEAK_NORMAL;
    else if ( st & (1<<(age5270_STAT_OSC_BIT-1))) *status = ILEAK_OSC;
    else if ( st & (1<<(age5270_STAT_OVF_BIT-1))) *status = ILEAK_OVERFLOW;
    else if ( st & (1<<(age5270_STAT_TCC_BIT-1))) *status = ILEAK_COMP;
    else if ( st & (1<<(age5270_STAT_OCC_BIT-1))) *status = ILEAK_COMP_OTHER;
    else if ( st & (1<<(age5270_STAT_SAB_BIT-1))) *status = ILEAK_TOO_SLOW;
    else if ( st & (1<<(age5270_STAT_SNF_BIT-1))) *status = ILEAK_TIMEOUT;
    else                                          *status = ILEAK_NORMAL;
    return( RETURN_PASS );
}

int error_info( int *errn, char *errm )
{
    int ret = 0;

    errn[0] = S_e5270.last_error;
    errn[1] = 0;
    if ( errm != NULL ) {
        if ( errn[0] == 0 ) {
            strcpy( errm, "" );
        }
        else {
            switch ( S_e5270.last_error_type ) {
            case ER_LIB  :
                switch ( errn[0] ) {
                case INVALID_PORT   : strcpy( errm, INVALID_PORT_MSG );   break;
                case INVALID_PIN    : strcpy( errm, INVALID_PIN_MSG );    break;
                case INVALID_RATIO  : strcpy( errm, INVALID_RATIO_MSG );  break;
                default             : strcpy( errm, "" );  break;
                }
                break;
            case ER_PNP  :
                ret = age5270_error_message( S_e5270.vi, errn[0], errm );
                CHECK_RET( ret, "error_info" );
                break;
            case ER_INST :
                strcpy( errm, S_e5270.inst_error_mes );
                break;
            }
        }
    }
    return( RETURN_PASS );
}

